
#ifndef DEBUG_INS_AND_OUTS
#define DEBUG_INS_AND_OUTS 0
#endif

/* for open(2) */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


/* for ndbm */
#include <db1/ndbm.h>

/* for rand */
#include <stdlib.h>

/* #include "slfprefs.h" -- passwd.c already includes this */

/* TODO: all the TODOs, and we need to put SLF_USER_DB into slfprefs.h */

/* public:
   slf_virtuser_get_info
   slf_virtuser_put_info
   slf_virtuser_remove
   slf_virtuser_getfirst
   slf_virtuser_getnext
   slf_virtuser_getstop
   slf_cryptnewpw
   slf_virtuser_get_dbfname
   slf_virtuser_set_dbfname
*/



/* for unpack */
static char from_db_usrname [32];
static char from_db_usrpw   [128];
static char from_db_usrgecos[250];
static struct passwd from_db_user_info;  /* has pointers to the above */

/* for pack */
static unsigned char packed_db_user_info[250];  /* TODO TODO TODO get max entry len for a ndbm record -- 1024 */


/* (from ndbm-solaris-2.6 manpage) 
 *     These database functions support  key/content  pairs  of  at
 * least 1024 bytes. */



/* for db file selection */
static const char *dbfname=NULL;

const char *slf_virtuser_get_dbfname(void) { return dbfname; }

void slf_virtuser_set_dbfname(const char *newfname)
{
  if ((newfname == NULL) || (newfname[0] == '\0'))
    dbfname = NULL;
  else
    dbfname = newfname;
}


#if DEBUG_INS_AND_OUTS
static void prnrec(datum dtu)
{
  int dlen = dtu.dsize;
  int i = 0;
  char *d = dtu.dptr;
  Lprintf ("\nsize is %d.", dlen);
  if (dlen > 160) {
    Lprintf ("** printing only first 160 bytes **");
    dlen = 160;
  }
  while (i < dlen)
  {
    // in 16-byte chunks...
    int j;
    char  *jp;
    Lprintf ("\n%4d [ ", i);
    for (j = 0, jp = d; (j+i<dlen)&&(j<16); ++j, ++jp)
    {
      if (j==8)
	Lprintf(" -");
      Lprintf (" %02x", (int) *jp);
    }
    Lprintf ("] ");
    for (j = 0, jp = d; (j+i<dlen)&&(j<16); ++j, ++jp)
      if ((*jp >= ' ') && (*jp <= (char)127))
	Lprintf ("%c", *jp);
      else
	Lprintf (".");

    i += 16;
    d += 16;
  }
  if (dlen < dtu.dsize)
    Lprintf("\n ** (%d more bytes not printed)", dtu.dsize - dlen);
  Lprintf ("\n\n");  
}
#endif /* DEBUG_INS_AND_OUTS */



/** given a buffer, pack the string "src" into it as a pascal string.
 *  Return a pointer to one past the last character of "src" in buf,
 *  or NULL if it won't fit or src or buf is NULL, or if src is greater
 *  than 255 chars long (unlikely in our application).
 */
static char* pack_pascal_string(unsigned char*buf, size_t bsiz, const char*src)
{
    size_t srclen;

    if ((buf==NULL) || (src==NULL))
	return NULL;

    srclen = strlen(src);
    if ((srclen >= (bsiz-1)) || (srclen > 255))
	return NULL;  /* won't fit (need room for length byte) */

    buf[0] = (unsigned char) srclen;   /* length byte */
    strncpy (&(buf[1]), src, srclen);  /* content */
    return &(buf[srclen + 1]);
}


/** given a packed pascal string, unpack it into destbuf as a C null-terminated
 *  string, and return a pointer to one past the end of that string, in the
 *  source buffer srcpstr.  Returns NULL if srcpstr=NULL,destbuf=NULL, or
 *  if it won't fit in destbuf.
 */
static char* unpack_pascal_string
    (unsigned char *srcpstr, char *destbuf, size_t destbsiz)
{
    size_t pstrsiz;

    if ((srcpstr==NULL)||(destbuf==NULL))
	return NULL;

    pstrsiz = (size_t) srcpstr[0];
    if (pstrsiz >= destbsiz)
	return NULL;  /* won't fit (need room for \0) */

    strncpy(destbuf, &(srcpstr[1]), pstrsiz);  /* content */
    destbuf[pstrsiz] = '\0';            /* null-terminate */
    return &(srcpstr[pstrsiz + 1]);
}




/* true for worked, false for didn't */
static bool passwd_struct_unpack
    (datum* rec, struct passwd *pswd, bool dynalloc)
{

  /* rec -> from_db_user_info */
  /* we set name, pw, gecos, gid (for admin or no) and leave the rest blank. */

  /* TODO TODO TODO currently assumes pristine, noncorrupted db record */
  unsigned char *c;
  c = (unsigned char *) (rec->dptr);  /* point c at start of db record */

#if DEBUG_INS_AND_OUTS
  Lprintf("L144: unpacking passwd from datum");
  prnrec(*rec);
#endif


  memset( pswd, 0, sizeof(struct passwd) );

  /* first byte is admin/nonadmin */
  if (((int) c[0]) == 1)
      pswd->pw_gid = admin_gid;
  else
      pswd->pw_gid = users_gid;
  ++c;

  /* name */
  c = unpack_pascal_string (c, from_db_usrname, sizeof(from_db_usrname));
  if (c == NULL)
      return false;  /* TODO TODO TODO error */

  /* passwd */
  c = unpack_pascal_string (c, from_db_usrpw, sizeof(from_db_usrpw));
  if (c == NULL)
      return false;  /* TODO TODO TODO error */


  /* gecos */
  c = unpack_pascal_string (c, from_db_usrgecos, sizeof(from_db_usrgecos));
  if (c == NULL)
      return false;


  /* now fill in the struct passwd */

  if (! dynalloc) {
    pswd->pw_name   = from_db_usrname;
    pswd->pw_passwd = from_db_usrpw;
    pswd->pw_gecos  = from_db_usrgecos;
  } else {
    pswd->pw_name   = strdup(from_db_usrname);
    pswd->pw_passwd = strdup(from_db_usrpw);
    pswd->pw_gecos  = strdup(from_db_usrgecos);
    if ((pswd->pw_name==NULL) || (pswd->pw_passwd==NULL)
	||(pswd->pw_gecos==NULL))
    {
      if (pswd->pw_name   !=NULL) free(pswd->pw_name);
      if (pswd->pw_passwd !=NULL) free(pswd->pw_passwd);
      if (pswd->pw_gecos  !=NULL) free(pswd->pw_gecos);
      return false;
    }
  }
  return true;
}


static void passwd_struct_pack
    (datum *rec, datum *key, const struct passwd *pack_from_here)
{
    /*  pack_from_here -> (rec,key) */
    /* out of:
          struct passwd {
	  char    *pw_name;
	  char    *pw_passwd;     
	  uid_t   pw_uid;         
	  gid_t   pw_gid;         
	  char    *pw_gecos;      
	  char    *pw_dir;        
	  char    *pw_shell;      
	  }; 
      we use name, passwd, gid, and gecos.
      I've decided the packing will be pascal strings, believe it or not.
      [length byte followed by content]
      if gid is admin_gid, the first byte in the record is 1, else it's 0.
    */
    unsigned char *c, *newc;
    size_t buflen;

    c = &(packed_db_user_info[0]);
    buflen = sizeof(packed_db_user_info);

    /* admin? */
    if (pack_from_here->pw_gid == admin_gid)
	c[0] = (unsigned char) 1;
    else
	c[0] = (unsigned char) 0;

    ++c;
    --buflen;

    /* name */
    newc = pack_pascal_string(c, buflen, pack_from_here->pw_name);
    if (newc == NULL)
	return;  /* TODO TODO TODO error */
    buflen -= (newc - c);
    c = newc;

    /* passwd */
    newc = pack_pascal_string(c, buflen, pack_from_here->pw_passwd);
    if (newc == NULL)
	return;  /* TODO TODO TODO error */
    buflen -= (newc - c);
    c = newc;

    /* gecos */
    newc = pack_pascal_string(c, buflen, pack_from_here->pw_gecos);
    if (newc == NULL)
	return;  /* TODO TODO TODO error */
    c = newc;
    
    /* datum */
    rec->dptr  = packed_db_user_info;
    rec->dsize = (int) ( c - &(packed_db_user_info[0]) );

    /* and key */
    key->dptr  = pack_from_here->pw_name;
    key->dsize = strlen(pack_from_here->pw_name);

#if DEBUG_INS_AND_OUTS
  Lprintf("L266: packing passwd into datum.  key:");
  prnrec(*key);
  Lprintf("L266: packing passwd into datum.  rec:");
  prnrec(*rec);
#endif
}


struct passwd *slf_virtuser_get_info
    (const char*name, const char*guess_of_pw, bool dynalloc)
{
    DBM* dbf;
    datum getthis, dbrec;
    struct passwd *pswd=NULL;

    if (dbfname == NULL)
	return NULL;   /* not running */
    if ((name==NULL) || (strlen(name)==0))
	return NULL;   /* TODO: "root"/"slforum"/"guest" */
    if ((guess_of_pw!=NULL) && (strlen(guess_of_pw)==0))
	return NULL;   /* they must have a pw */

    /** flags, mode are same meaning as in open(2). */
    dbf = dbm_open(dbfname, O_RDONLY, 0700);
    if (dbf==NULL)
	return NULL;  /* TODO: error msg */

    getthis.dptr  = (char *) name;
    getthis.dsize = strlen(name);  /* TODO cache name strlen */

#if DEBUG_INS_AND_OUTS
    Lprintf ("L306: fetching record with key:\n");
    prnrec (getthis);
#endif

    dbrec = dbm_fetch (dbf, getthis);

    if (dbrec.dptr == NULL) {
#if DEBUG_INS_AND_OUTS
	Lprintf("L313: got NULL\n");
#endif
	return NULL;  /* not in DB */
    }

#if DEBUG_INS_AND_OUTS
    Lprintf ("L317: got datum:\n");
    prnrec (dbrec);
#endif

    /* copy fields from dptr to from_db_user_info after verifying size */
    if (!dynalloc)
      pswd = &from_db_user_info;
    else {
      pswd = (struct passwd *)malloc(sizeof(struct passwd));
      if (pswd==NULL) {
	// TODO TODO TODO can we announce anything?
	dbm_close(dbf);
	return NULL;
      }
    }
    if (!passwd_struct_unpack(&dbrec, pswd, dynalloc)) {
      // TODO TODO TODO announce?
      if(dynalloc) free(pswd);
      dbm_close(dbf);
      return NULL;
    }

    dbm_close(dbf);

    /* check password */
    if (guess_of_pw != NULL)
    {
      char salt[3];
      const char* crypted;

      salt[0] = pswd->pw_passwd[0];
      salt[1] = pswd->pw_passwd[1];
      salt[2] = '\0';
      crypted = crypt(guess_of_pw, salt);
	
      if (0 != strcmp(pswd->pw_passwd, crypted))
	return NULL;  /* pw fails to match */
    }

    return pswd;   /* they got it */
}



bool slf_virtuser_put_info (const struct passwd *udata)
{
    DBM* dbf;
    datum putthis, putkey;

    if (dbfname==NULL)
	return false;
    if (udata==NULL)
	return false;


    if (strlen(udata->pw_passwd) < 4)
	return false;       /* must have a pw */

    passwd_struct_pack(&putthis, &putkey, udata);

    dbf = dbm_open(dbfname, O_RDWR|O_CREAT, 0700);
    if (dbf==NULL)
	return false;  /* TODO: error msg */

    if (0 > dbm_store (dbf, putkey, putthis, DBM_REPLACE)) {
      dbm_close(dbf);
      return false;
    }
    dbm_close(dbf);
    return true;
}



bool slf_virtuser_remove (const char *uname)
{
    DBM* dbf;
    datum delkey;

    if (dbfname == NULL)
	return false;
    if ((uname==NULL)||(uname[0] == '\0'))
	return true;  /* how can that fail? */

    delkey.dptr = (char *) uname;
    delkey.dsize = strlen(uname);

    dbf = dbm_open(dbfname, O_RDWR|O_CREAT, 0700);
    if (dbf==NULL)
	return false;  /* TODO: error msg */

    if (0 != dbm_delete (dbf, delkey)) {    /* retval: 0 success, neg fail */
      /* TODO TODO TODO error msg if not success? */
      dbm_close(dbf);
      return false;
    }

    dbm_close (dbf);
    return true;
}



static DBM* gfn_dbf = NULL;   /* for getfirstnext */

const struct passwd *slf_virtuser_getfirst(void)
{
  datum dbrec;

  if(dbfname==NULL)
    return NULL;

    /** flags, mode are same meaning as in open(2). */
    gfn_dbf = dbm_open(dbfname, O_RDONLY, 0700);
    if (gfn_dbf==NULL)
	return NULL;  /* TODO: error msg */

    dbrec = dbm_firstkey (gfn_dbf);
    if (dbrec.dptr == NULL) {
#if DEBUG_INS_AND_OUTS
	Lprintf("L431: firstkey got NULL\n");
#endif
	return NULL;  /* DB empty */
    }

#if DEBUG_INS_AND_OUTS
    Lprintf("L433: firstkey got:\n");
    prnrec(dbrec);
#endif

    dbrec = dbm_fetch(gfn_dbf, dbrec);  /* get value for this key */
    if (dbrec.dptr == NULL)
      return NULL;  /* empty */

    /* copy fields from dptr to from_db_user_info after verifying size */
    passwd_struct_unpack(&dbrec, &from_db_user_info, false);
    return &from_db_user_info;
}


const struct passwd *slf_virtuser_getnext(void)
{
  datum dbrec;

  if (gfn_dbf == NULL)
    return NULL;  /* error msg? */

  dbrec = dbm_nextkey (gfn_dbf);
  if (dbrec.dptr == NULL) {
#if DEBUG_INS_AND_OUTS
    Lprintf("L455: nextkey got NULL\n");
#endif
    return NULL;  /* DB empty */
  }

#if DEBUG_INS_AND_OUTS
  Lprintf("L457: nextkey got:\n");
  prnrec(dbrec);
#endif

  dbrec = dbm_fetch(gfn_dbf, dbrec);  /* get value for this key */
  if (dbrec.dptr == NULL)
    return NULL;  /* empty */

  /* copy fields from dptr to from_db_user_info after verifying size */
  passwd_struct_unpack(&dbrec, &from_db_user_info, false);
  return &from_db_user_info;
}


void slf_virtuser_getstop(void)
{
  if (gfn_dbf != NULL)
    dbm_close(gfn_dbf);
  gfn_dbf = NULL;
}


const char *slf_cryptnewpw (const char *unenc_pw)
{
  static const char* saltstr
    = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
  char salt[3];

  salt[0] = saltstr[rand() % 64];
  salt[1] = saltstr[rand() % 64];
  salt[2] = '\0';

  return crypt (unenc_pw, salt);
}


